"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[8225],{43598:function(t,e,a){a.r(e),a.d(e,{icon:function(){return l}});a(72791);var r=a(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var a=arguments[e];for(var r in a)Object.prototype.hasOwnProperty.call(a,r)&&(t[r]=a[r])}return t},n.apply(this,arguments)}function h(t,e){if(null==t)return{};var a,r,n=function(t,e){if(null==t)return{};var a,r,n={},h=Object.keys(t);for(r=0;r<h.length;r++)a=h[r],e.indexOf(a)>=0||(n[a]=t[a]);return n}(t,e);if(Object.getOwnPropertySymbols){var h=Object.getOwnPropertySymbols(t);for(r=0;r<h.length;r++)a=h[r],e.indexOf(a)>=0||Object.prototype.propertyIsEnumerable.call(t,a)&&(n[a]=t[a])}return n}var l=function(t){var e=t.title,a=t.titleId,l=h(t,["title","titleId"]);return(0,r.tZ)("svg",n({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":a},l),e?(0,r.tZ)("title",{id:a},e):null,(0,r.tZ)("path",{d:"M29 12a3 3 0 00-3 3h-4a3 3 0 00-3-3h-6a3 3 0 00-3 3H6a3 3 0 00-3-3H0v14h3a3 3 0 003-3h4a3 3 0 003 3h6a3 3 0 003-3h4a3 3 0 003 3h3V12h-3zM3 24H2V14h1a1 1 0 011 1v8a1 1 0 01-1 1zm17-3v2a1 1 0 01-1 1h-6a1 1 0 01-1-1v-2H6v-4h6v-2a1 1 0 011-1h6a1 1 0 011 1v2h6v4h-6zm10 3h-1a1 1 0 01-1-1v-8a1 1 0 011-1h1v10z"}),(0,r.tZ)("path",{className:"euiIcon__fillSecondary",d:"M22 6H10v2h5v2h2V8h5z"}))}}}]);
//# sourceMappingURL=icon.app_pipeline.be125910.chunk.js.map