"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[2067],{31201:function(t,e,l){l.r(e),l.d(e,{icon:function(){return a}});l(72791);var r=l(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var r in l)Object.prototype.hasOwnProperty.call(l,r)&&(t[r]=l[r])}return t},n.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,r,n=function(t,e){if(null==t)return{};var l,r,n={},i=Object.keys(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||(n[l]=t[l]);return n}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(n[l]=t[l])}return n}var a=function(t){var e=t.title,l=t.titleId,a=i(t,["title","titleId"]);return(0,r.tZ)("svg",n({width:32,height:32,viewBox:"0 0 32 32",xmlns:"http://www.w3.org/2000/svg","aria-labelledby":l},a),e?(0,r.tZ)("title",{id:l},e):null,(0,r.tZ)("path",{d:"M2.996 18.636L0 20.434v6.132l6 3.6 5-3 5 3 5-3 5 3 6-3.6v-6.132l-5-3v-5l-5-3v-5l-6-3.6-6 3.6v2.832l2 .033V5.566l4-2.4 4 2.4v3.868l-1.996 1.197 1 1.733L21 11.166l4 2.4v3.868l-4 2.4-2.257-1.354-.971 1.75L20 21.565v3.868l-4 2.4-4-2.4v-1.7l-2-.033v1.733l-4 2.4-4-2.4v-3.868l1.996-1.197-1-1.733zM22 25.434v-3.868l4-2.4 4 2.4v3.868l-4 2.4-4-2.4z"}),(0,r.tZ)("path",{className:"euiIcon__fillSecondary",d:"M11 22.166l-6-3.6v-6.132l6-3.6 6 3.6v6.132l-6 3.6zm4-4.732v-3.868l-4-2.4-4 2.4v3.868l4 2.4 4-2.4z"}))}}}]);
//# sourceMappingURL=icon.app_fleet.cd29c517.chunk.js.map