"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[5331],{78671:function(t,e,r){r.r(e),r.d(e,{icon:function(){return i}});r(72791);var n=r(52554);function h(){return h=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t},h.apply(this,arguments)}function a(t,e){if(null==t)return{};var r,n,h=function(t,e){if(null==t)return{};var r,n,h={},a=Object.keys(t);for(n=0;n<a.length;n++)r=a[n],e.indexOf(r)>=0||(h[r]=t[r]);return h}(t,e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(t);for(n=0;n<a.length;n++)r=a[n],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(h[r]=t[r])}return h}var i=function(t){var e=t.title,r=t.titleId,i=a(t,["title","titleId"]);return(0,n.tZ)("svg",h({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},i),e?(0,n.tZ)("title",{id:r},e):null,(0,n.tZ)("path",{d:"M4 3a1 1 0 00-1 1v8a1 1 0 001 1h8a1 1 0 001-1V4a1 1 0 00-1-1H4zm10 8v1a2 2 0 01-2 2h-1v2h-1v-2H8.5v2h-1v-2H6v2H5v-2H4a2 2 0 01-2-2v-1H0v-1h2V8.5H0v-1h2V6H0V5h2V4a2 2 0 012-2h1V0h1v2h1.5V0h1v2H10V0h1v2h1a2 2 0 012 2v1h2v1h-2v1.5h2v1h-2V10h2v1h-2z"}),(0,n.tZ)("rect",{width:6,height:6,x:5,y:5,rx:1}))}}}]);
//# sourceMappingURL=icon.compute.3a55b39e.chunk.js.map