"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[4732],{35300:function(t,e,l){l.r(e),l.d(e,{icon:function(){return u}});l(72791);var r=l(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var r in l)Object.prototype.hasOwnProperty.call(l,r)&&(t[r]=l[r])}return t},n.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,r,n=function(t,e){if(null==t)return{};var l,r,n={},i=Object.keys(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||(n[l]=t[l]);return n}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(n[l]=t[l])}return n}var u=function(t){var e=t.title,l=t.titleId,u=i(t,["title","titleId"]);return(0,r.tZ)("svg",n({width:16,height:16,viewBox:"0 0 16 16",xmlns:"http://www.w3.org/2000/svg","aria-labelledby":l},u),e?(0,r.tZ)("title",{id:l},e):null,(0,r.tZ)("path",{d:"M7.808 10.197H6.796L5.859 13H4.485l.937-2.803H3.966l.219-1.25h1.647l.608-1.805H4.991l.226-1.251h1.64l.95-2.844h1.368l-.95 2.844h1.018l.95-2.844h1.374l-.95 2.844h1.51l-.218 1.25h-1.702l-.608 1.805h1.497l-.219 1.251H9.182L8.252 13H6.878l.93-2.803zm-.602-1.25h1.012l.615-1.805H7.814l-.608 1.804z",fillRule:"evenodd"}))}}}]);
//# sourceMappingURL=icon.number.dd0b0c9e.chunk.js.map