syntax = "proto3";

package grpc.health.v1;

option java_package = "io.grpc.health.v1";
option java_outer_classname = "HealthProto";

message HealthCheckRequest {
  string service = 1;
}

enum ServingStatus {
  UNKNOWN = 0;
  SERVING = 1;
  NOT_SERVING = 2;
}

message HealthCheckResponse {
  ServingStatus status = 1;
}

service Health {
  rpc Check(HealthCheckRequest) returns (HealthCheckResponse);
}