"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[6440],{63352:function(t,e,l){l.r(e),l.d(e,{icon:function(){return a}});l(72791);var n=l(52554);function r(){return r=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var n in l)Object.prototype.hasOwnProperty.call(l,n)&&(t[n]=l[n])}return t},r.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,n,r=function(t,e){if(null==t)return{};var l,n,r={},i=Object.keys(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||(r[l]=t[l]);return r}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(r[l]=t[l])}return r}var a=function(t){var e=t.title,l=t.titleId,a=i(t,["title","titleId"]);return(0,n.tZ)("svg",r({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":l},a),e?(0,n.tZ)("title",{id:l},e):null,(0,n.tZ)("g",{fill:"none",fillRule:"evenodd"},(0,n.tZ)("path",{fill:"#119639",d:"M16 0l13.856 8v16L16 32 2.144 24V8z"}),(0,n.tZ)("path",{fill:"#FFF",fillRule:"nonzero",d:"M11.17 13.512v8.376a1.607 1.607 0 11-3.215 0V9.632c0-1.432 1.731-2.149 2.744-1.136l9.51 9.512V9.632a1.607 1.607 0 013.215 0v12.256c0 1.432-1.731 2.149-2.744 1.136l-9.51-9.512z"})))}}}]);
//# sourceMappingURL=icon.logo_nginx.0ac8f945.chunk.js.map