"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[2236],{76562:function(t,e,l){l.r(e),l.d(e,{icon:function(){return a}});l(72791);var n=l(52554);function r(){return r=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var n in l)Object.prototype.hasOwnProperty.call(l,n)&&(t[n]=l[n])}return t},r.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,n,r=function(t,e){if(null==t)return{};var l,n,r={},i=Object.keys(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||(r[l]=t[l]);return r}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(r[l]=t[l])}return r}var a=function(t){var e=t.title,l=t.titleId,a=i(t,["title","titleId"]);return(0,n.tZ)("svg",r({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":l},a),e?(0,n.tZ)("title",{id:l},e):null,(0,n.tZ)("g",{fill:"none",fillRule:"evenodd"},(0,n.tZ)("path",{fill:"#F04E98",d:"M2 32h28V20l-6.465-6.465a5 5 0 00-7.07 0L2 28v4z"}),(0,n.tZ)("path",{className:"euiIcon__fillNegative",d:"M16.465 13.535l-3.536 3.536a9.965 9.965 0 007.07 2.93 9.965 9.965 0 007.072-2.93l-3.536-3.536a5 5 0 00-7.07 0"}),(0,n.tZ)("path",{fill:"#FEC514",d:"M14.343 11.414A7.951 7.951 0 0120 9.071c2.137 0 4.146.832 5.657 2.343l3.207 3.207A9.955 9.955 0 0030 10.001c0-5.524-4.477-10-10-10-5.522 0-10 4.476-10 10 0 1.667.414 3.237 1.137 4.62l3.206-3.207z"})))}}}]);
//# sourceMappingURL=icon.logo_metrics.f7e79516.chunk.js.map