"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[8473],{2357:function(t,e,r){r.r(e),r.d(e,{icon:function(){return l}});r(72791);var n=r(52554);function h(){return h=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t},h.apply(this,arguments)}function v(t,e){if(null==t)return{};var r,n,h=function(t,e){if(null==t)return{};var r,n,h={},v=Object.keys(t);for(n=0;n<v.length;n++)r=v[n],e.indexOf(r)>=0||(h[r]=t[r]);return h}(t,e);if(Object.getOwnPropertySymbols){var v=Object.getOwnPropertySymbols(t);for(n=0;n<v.length;n++)r=v[n],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(h[r]=t[r])}return h}var l=function(t){var e=t.title,r=t.titleId,l=v(t,["title","titleId"]);return(0,n.tZ)("svg",h({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},l),e?(0,n.tZ)("title",{id:r},e):null,(0,n.tZ)("path",{d:"M1 5V1h4v4H1zm3-1V2H2v2h2zm2 1V1h4v4H6zm3-1V2H7v2h2zm2 1V1h4v4h-4zm1-1h2V2h-2v2zM1 10V6h4v4H1zm3-1V7H2v2h2zm2 1V6h4v4H6zm3-1V7H7v2h2zm2 1V6h4v4h-4zm3-1V7h-2v2h2zM1 15v-4h4v4H1zm1-1h2v-2H2v2zm4 1v-4h4v4H6zm1-1h2v-2H7v2zm4 1v-4h4v4h-4zm1-1h2v-2h-2v2z"}))}}}]);
//# sourceMappingURL=icon.grid.0e211169.chunk.js.map