"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[5177],{94746:function(t,e,r){r.r(e),r.d(e,{icon:function(){return a}});r(72791);var l=r(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var l in r)Object.prototype.hasOwnProperty.call(r,l)&&(t[l]=r[l])}return t},n.apply(this,arguments)}function i(t,e){if(null==t)return{};var r,l,n=function(t,e){if(null==t)return{};var r,l,n={},i=Object.keys(t);for(l=0;l<i.length;l++)r=i[l],e.indexOf(r)>=0||(n[r]=t[r]);return n}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(l=0;l<i.length;l++)r=i[l],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(n[r]=t[r])}return n}var a=function(t){var e=t.title,r=t.titleId,a=i(t,["title","titleId"]);return(0,l.tZ)("svg",n({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},a),e?(0,l.tZ)("title",{id:r},e):null,(0,l.tZ)("path",{d:"M7.8 9.812h-.842l-.78 2.335H5.031l.78-2.335H4.6l.182-1.043h1.373l.507-1.504H5.454l.188-1.042h1.367l.792-2.37H8.94l-.792 2.37h.849l.792-2.37h1.145l-.792 2.37H11.4l-.182 1.042H9.8L9.293 8.77h1.248l-.183 1.043H8.946l-.775 2.335H7.026L7.8 9.812zm-.5-1.043h.842l.513-1.504h-.849L7.3 8.77z"}))}}}]);
//# sourceMappingURL=icon.tokenNumber.892c6f00.chunk.js.map