"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[3511],{70297:function(t,e,r){r.r(e),r.d(e,{icon:function(){return a}});r(72791);var l=r(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var l in r)Object.prototype.hasOwnProperty.call(r,l)&&(t[l]=r[l])}return t},n.apply(this,arguments)}function i(t,e){if(null==t)return{};var r,l,n=function(t,e){if(null==t)return{};var r,l,n={},i=Object.keys(t);for(l=0;l<i.length;l++)r=i[l],e.indexOf(r)>=0||(n[r]=t[r]);return n}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(l=0;l<i.length;l++)r=i[l],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(n[r]=t[r])}return n}var a=function(t){var e=t.title,r=t.titleId,a=i(t,["title","titleId"]);return(0,l.tZ)("svg",n({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":r},a),e?(0,l.tZ)("title",{id:r},e):null,(0,l.tZ)("g",{fill:"none",fillRule:"evenodd"},(0,l.tZ)("path",{fill:"#F04E98",d:"M32 9V2.5A2.5 2.5 0 0029.5 0h-27A2.5 2.5 0 000 2.5V9h32z"}),(0,l.tZ)("path",{fill:"#00BFB3",d:"M0 20h32v-8H0z"}),(0,l.tZ)("path",{fill:"#0080D5",d:"M14.5 23H0v6.5A2.5 2.5 0 002.5 32h12v-9z"}),(0,l.tZ)("path",{fill:"#FEC514",d:"M17.5 23v9h12a2.5 2.5 0 002.5-2.5V23H17.5z"})))}}}]);
//# sourceMappingURL=icon.logo_elastic_stack.23cf6b4e.chunk.js.map