"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[7833],{44060:function(e,t,l){l.r(t),l.d(t,{icon:function(){return i}});l(72791);var n=l(52554);function r(){return r=Object.assign||function(e){for(var t=1;t<arguments.length;t++){var l=arguments[t];for(var n in l)Object.prototype.hasOwnProperty.call(l,n)&&(e[n]=l[n])}return e},r.apply(this,arguments)}function a(e,t){if(null==e)return{};var l,n,r=function(e,t){if(null==e)return{};var l,n,r={},a=Object.keys(e);for(n=0;n<a.length;n++)l=a[n],t.indexOf(l)>=0||(r[l]=e[l]);return r}(e,t);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);for(n=0;n<a.length;n++)l=a[n],t.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(e,l)&&(r[l]=e[l])}return r}var i=function(e){var t=e.title,l=e.titleId,i=a(e,["title","titleId"]);return(0,n.tZ)("svg",r({width:32,height:32,viewBox:"0 0 32 32",xmlns:"http://www.w3.org/2000/svg","aria-labelledby":l},i),t?(0,n.tZ)("title",{id:l},t):null,(0,n.tZ)("path",{className:"euiIcon__fillSecondary",d:"M21 2.82L16 .038 11 2.82v2.289l5-2.782 5 2.782v-2.29z"}),(0,n.tZ)("path",{className:"euiIcon__fillSecondary",d:"M21 7.282L16 4.5l-5 2.782V9.57l5-2.781 5 2.781V7.282z"}),(0,n.tZ)("path",{d:"M7 5.045L2 7.827v15.577l14 7.788 14-7.788V7.827l-5-2.782v2.289l3 1.669v13.225l-12 6.676-12-6.676V9.003l3-1.669V5.045z"}),(0,n.tZ)("path",{className:"euiIcon__fillSecondary",fillRule:"evenodd",clipRule:"evenodd",d:"M22 12.5L16 9l-6 3.5v7l6 3.5 6-3.5v-7zm-9.974 1.205L16 11.387l3.974 2.318v4.59L16 20.613l-3.974-2.318v-4.59z"}))}}}]);
//# sourceMappingURL=icon.app_agent.3fc0b866.chunk.js.map