//
// * Copyright 2021 The Feast Authors
// *
// * Licensed under the Apache License, Version 2.0 (the "License");
// * you may not use this file except in compliance with the License.
// * You may obtain a copy of the License at
// *
// *     https://www.apache.org/licenses/LICENSE-2.0
// *
// * Unless required by applicable law or agreed to in writing, software
// * distributed under the License is distributed on an "AS IS" BASIS,
// * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// * See the License for the specific language governing permissions and
// * limitations under the License.
//

syntax = "proto3";

package feast.core;
option java_package = "feast.proto.core";
option java_outer_classname = "DatastoreTableProto";
option go_package = "github.com/feast-dev/feast/go/protos/feast/core";

import "google/protobuf/wrappers.proto";

// Represents a Datastore table
message DatastoreTable {
    // Feast project of the table
    string project = 1;

    // Name of the table
    string name = 2;

    // GCP project id
    google.protobuf.StringValue project_id = 3;

    // Datastore namespace
    google.protobuf.StringValue namespace = 4;
}