"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[7466],{98310:function(t,e,n){n.r(e),n.d(e,{icon:function(){return a}});n(72791);var r=n(52554);function l(){return l=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(t[r]=n[r])}return t},l.apply(this,arguments)}function i(t,e){if(null==t)return{};var n,r,l=function(t,e){if(null==t)return{};var n,r,l={},i=Object.keys(t);for(r=0;r<i.length;r++)n=i[r],e.indexOf(n)>=0||(l[n]=t[n]);return l}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(r=0;r<i.length;r++)n=i[r],e.indexOf(n)>=0||Object.prototype.propertyIsEnumerable.call(t,n)&&(l[n]=t[n])}return l}var a=function(t){var e=t.title,n=t.titleId,a=i(t,["title","titleId"]);return(0,r.tZ)("svg",l({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":n},a),e?(0,r.tZ)("title",{id:n},e):null,(0,r.tZ)("g",{fill:"none",fillRule:"evenodd"},(0,r.tZ)("path",{fill:"#00BFB3",d:"M0 22c0 5.522 4.478 10 10 10V12C4.478 12 0 16.478 0 22"}),(0,r.tZ)("path",{className:"euiIcon__fillNegative",d:"M10 12v10h10c0-5.522-4.478-10-10-10"}),(0,r.tZ)("path",{fill:"#F04E98",d:"M10 0v9c7.168 0 13 5.832 13 13h9C32 9.85 22.15 0 10 0"})))}}}]);
//# sourceMappingURL=icon.logo_business_analytics.25d7b2b2.chunk.js.map