"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[1403],{36020:function(t,e,r){r.r(e),r.d(e,{icon:function(){return u}});r(72791);var n=r(52554);function i(){return i=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t},i.apply(this,arguments)}function l(t,e){if(null==t)return{};var r,n,i=function(t,e){if(null==t)return{};var r,n,i={},l=Object.keys(t);for(n=0;n<l.length;n++)r=l[n],e.indexOf(r)>=0||(i[r]=t[r]);return i}(t,e);if(Object.getOwnPropertySymbols){var l=Object.getOwnPropertySymbols(t);for(n=0;n<l.length;n++)r=l[n],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(i[r]=t[r])}return i}var u=function(t){var e=t.title,r=t.titleId,u=l(t,["title","titleId"]);return(0,n.tZ)("svg",i({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},u),e?(0,n.tZ)("title",{id:r},e):null,(0,n.tZ)("rect",{width:10,height:1.5,x:3,y:7.25,rx:.5}))}}}]);
//# sourceMappingURL=icon.minus.736b4bf4.chunk.js.map