"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[5434],{23690:function(t,e,r){r.r(e),r.d(e,{icon:function(){return i}});r(72791);var n=r(52554);function l(){return l=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t},l.apply(this,arguments)}function h(t,e){if(null==t)return{};var r,n,l=function(t,e){if(null==t)return{};var r,n,l={},h=Object.keys(t);for(n=0;n<h.length;n++)r=h[n],e.indexOf(r)>=0||(l[r]=t[r]);return l}(t,e);if(Object.getOwnPropertySymbols){var h=Object.getOwnPropertySymbols(t);for(n=0;n<h.length;n++)r=h[n],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(l[r]=t[r])}return l}var i=function(t){var e=t.title,r=t.titleId,i=h(t,["title","titleId"]);return(0,n.tZ)("svg",l({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},i),e?(0,n.tZ)("title",{id:r},e):null,(0,n.tZ)("path",{fillRule:"evenodd",d:"M13 10v3h-3v-1H6v1H3v-3h1V6H3V3h3v1h4V3h3v3h-1v4h1zm-8 1H4v1h1v-1zm7 0h-1v1h1v-1zM5 4H4v1h1V4zm7 0h-1v1h1V4zm-1 2h-1V5H6v1H5v4h1v1h4v-1h1V6z"}))}}}]);
//# sourceMappingURL=icon.tokenShape.0d3d5650.chunk.js.map