"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[5451],{66513:function(t,e,l){l.r(e),l.d(e,{icon:function(){return a}});l(72791);var r=l(52554);function n(){return n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var r in l)Object.prototype.hasOwnProperty.call(l,r)&&(t[r]=l[r])}return t},n.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,r,n=function(t,e){if(null==t)return{};var l,r,n={},i=Object.keys(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||(n[l]=t[l]);return n}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(r=0;r<i.length;r++)l=i[r],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(n[l]=t[l])}return n}var a=function(t){var e=t.title,l=t.titleId,a=i(t,["title","titleId"]);return(0,r.tZ)("svg",n({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":l},a),e?(0,r.tZ)("title",{id:l},e):null,(0,r.tZ)("g",{fill:"none"},(0,r.tZ)("path",{fill:"#FFAE00",d:"M16 32L0 11.454l6.931-10.38L16 0l9.069 1.074L32 11.454z"}),(0,r.tZ)("path",{fill:"#EC6C00",d:"M16 32L0 11.454h32z"}),(0,r.tZ)("path",{fill:"#FFAE00",d:"M16 32L6.477 11.454h19.045z"}),(0,r.tZ)("path",{fill:"#FFEFB4",d:"M16 0L6.477 11.454h19.045z"}),(0,r.tZ)("path",{fill:"#FFAE00",d:"M6.932 1.074L3.369 6.3.001 11.454h6.542zM25.069 1.074L28.632 6.3 32 11.454h-6.542z"}),(0,r.tZ)("path",{fill:"#FED305",d:"M6.931 1.074l-.453 10.38L16 0zM25.069 1.074l.453 10.38L16 0z"})))}}}]);
//# sourceMappingURL=icon.logo_sketch.0220dad0.chunk.js.map