"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[467],{3111:function(t,e,r){r.r(e),r.d(e,{icon:function(){return l}});r(72791);var n=r(52554);function a(){return a=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t},a.apply(this,arguments)}function h(t,e){if(null==t)return{};var r,n,a=function(t,e){if(null==t)return{};var r,n,a={},h=Object.keys(t);for(n=0;n<h.length;n++)r=h[n],e.indexOf(r)>=0||(a[r]=t[r]);return a}(t,e);if(Object.getOwnPropertySymbols){var h=Object.getOwnPropertySymbols(t);for(n=0;n<h.length;n++)r=h[n],e.indexOf(r)>=0||Object.prototype.propertyIsEnumerable.call(t,r)&&(a[r]=t[r])}return a}var l=function(t){var e=t.title,r=t.titleId,l=h(t,["title","titleId"]);return(0,n.tZ)("svg",a({xmlns:"http://www.w3.org/2000/svg",width:16,height:16,viewBox:"0 0 16 16","aria-labelledby":r},l),e?(0,n.tZ)("title",{id:r},e):null,(0,n.tZ)("path",{d:"M2 2.729V2a1 1 0 011-1h2v1H3v12h4v1H3a1 1 0 01-1-1V2.729zM14 5V2a1 1 0 00-1-1h-2v1h2v3h1zm-1 1h2v9H8V6h5V5H8a1 1 0 00-1 1v9a1 1 0 001 1h7a1 1 0 001-1V6a1 1 0 00-1-1h-2v1z"}),(0,n.tZ)("path",{d:"M9 10h5V9H9v1zm0-2h5V7H9v1zm0 4h5v-1H9v1zm0 2h5v-1H9v1zm2-12V1a1 1 0 00-1-1H6a1 1 0 00-1 1v1h1V1h4v1h1zM5 3h6V2H5v1z"}))}}}]);
//# sourceMappingURL=icon.copy_clipboard.b4f319c9.chunk.js.map