"use strict";(self.webpackChunkexample_feast_ui=self.webpackChunkexample_feast_ui||[]).push([[5110],{67486:function(t,e,l){l.r(e),l.d(e,{icon:function(){return a}});l(72791);var n=l(52554);function r(){return r=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var l=arguments[e];for(var n in l)Object.prototype.hasOwnProperty.call(l,n)&&(t[n]=l[n])}return t},r.apply(this,arguments)}function i(t,e){if(null==t)return{};var l,n,r=function(t,e){if(null==t)return{};var l,n,r={},i=Object.keys(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||(r[l]=t[l]);return r}(t,e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);for(n=0;n<i.length;n++)l=i[n],e.indexOf(l)>=0||Object.prototype.propertyIsEnumerable.call(t,l)&&(r[l]=t[l])}return r}var a=function(t){var e=t.title,l=t.titleId,a=i(t,["title","titleId"]);return(0,n.tZ)("svg",r({xmlns:"http://www.w3.org/2000/svg",width:32,height:32,viewBox:"0 0 32 32","aria-labelledby":l},a),e?(0,n.tZ)("title",{id:l},e):null,(0,n.tZ)("g",{fill:"none",fillRule:"evenodd"},(0,n.tZ)("path",{fill:"#F04E98",d:"M4 0v28.789L28.935.017z"}),(0,n.tZ)("path",{className:"euiIcon__fillNegative",d:"M4 12v16.789l11.906-13.738A24.721 24.721 0 004 12"}),(0,n.tZ)("path",{fill:"#00BFB3",d:"M18.479 16.664L6.268 30.754l-1.073 1.237h23.191c-1.252-6.292-4.883-11.719-9.908-15.327"})))}}}]);
//# sourceMappingURL=icon.logo_kibana.bae5f9b7.chunk.js.map